%{
#include <stdio.h>
#include <string.h>

extern int yyparse ();

extern "C" int yywrap () { return 1; }

#include "parserincludes.h"

extern Program *AlluvionProgram;

#define YYERROR_VERBOSE

%}

%union
{
  int                Number;
  char               *String;
  Expression         *Expr;
  VectorExpression   *VExpr;
  ArgumentExpression *Arg;
  RangeExpression    *Rng;
  RangesExpression   *Rngs;
  ArgumentVector     *ArgVec;
  Boolean            *Bool;
  Statement          *Stmnt;
  Sequence           *Seq;
  Function           *Fun;
  Program            *Prog;
}

%token <String> NAME
%token <Number> NUM
%token <String> BITS
%token FOR IF ELSE WHILE PRINT
%token '=' '>' '<' GE LE
%left '&' '|'
%left NOT
%left '+' '-'
%left '*' '/' '%'
%left NEG
%right '^'

%type <Expr>   expression
%type <Expr>   expressions
%type <VExpr>  vectorexpression
%type <Expr>   controlexpression
%type <ArgVec> args
%type <ArgVec> argvec
%type <Stmnt>  statement
%type <Bool>   boolean
%type <Seq>    sequence
%type <Seq>    body
%type <Fun>    function
%type <Arg>    arg
%type <Rng>    range
%type <Rngs>   ranges

//%error-verbose
%locations

%{

extern int yylex ();

extern YYLTYPE yylloc;

void yyerror (const char *str)
{
  cerr << "ERROR at " << yylloc << ": " << str << endl;
}

%}

%%

program:
| program function
  {
    AlluvionProgram->AddFunction ($2);
  };

function:
  args '=' NAME args body
  {
    $$ = new Function ($3, $1, $4, $5);
  }
| args '=' NAME '<' argvec '>' args body
  {
    $$ = new Function ($3, $1, $7, $8, $5);
  };

args:
  arg
  {
    $$ = new ArgumentVector ($1);
    $$->SetLocation (@$);
  } 
| '(' argvec ')'
  {
    $$ = $2;
    $$->SetLocation (@$);
  };

argvec:
  arg
  {
    $$ = new ArgumentVector ($1);
  }
| argvec ',' arg
  {
    $1->AddArgument ($3);
    $$ = $1;
  };

arg:
  NAME
  {
    $$ = new ArgumentExpression ($1, @$);
  }
| NAME '[' ranges ']'
  {
    $$ = new ArgumentExpression ($1, $3, @$);
  };

range:
  controlexpression
  {
    $$ = new RangeExpression ($1);
    $$->SetLocation (@$);
  }
| controlexpression ':' controlexpression
  {
    $$ = new RangeExpression ($1, $3);
    $$->SetLocation (@$);
  }
| controlexpression ':' controlexpression ':' controlexpression
  {
    $$ = new RangeExpression ($1, $3, $5);
    $$->SetLocation (@$);
  };

ranges:
  range
  {
     $$ = new RangesExpression ();
     $$->AddRange ($1);
  }
| ranges ',' range
  {
     $$->AddRange ($3);
  };

body:
  '{' sequence '}'
  {
    $$ = $2;
  };

sequence:
  {
    $$ = new Sequence ();
  }
| sequence statement
  {
    $1->AddStatement ($2);
  };

statement:
  args '=' expression ';'
  { $$ = new Assignment ($1, $3, @$); }
| FOR NAME '=' '[' range ']' body
  { $$ = new ForStatement ($2, $5, $7, @$); }
| IF boolean body
  { $$ = new IfStatement ($2, $3, @$); }
| IF boolean body ELSE body
  { $$ = new IfStatement ($2, $3, $5, @$); }
| WHILE boolean body
  { $$ = new WhileStatement ($2, $3, @$); }
| PRINT args ';'
  { $$ = new PrintStatement ($2, @$); };

expression:
  NAME
  { $$ = new ArgumentExpression ($1, @$); }
| NAME '[' ranges ']'
  { $$ = new ArgumentExpression ($1, $3, @$); }
| '(' expressions ')'
  { $2->SetLocation (@$); $$ = $2; }
| '(' ')'
  { $$ = new EmptyExpression (@$); }
| NAME '(' expressions ')'
  { $$ = new CallExpression ($1, $3, @$); }
| NAME '<' expressions '>' '(' expressions ')'
  { $$ = new CallExpression ($1, $6, $3, @$); }
| BITS
  { $$ = new BitExpression ($1, @$); }
| NUM
  { $$ = new NumberExpression ($1, @$); }
| expression '+' expression
  { $$ = new BinaryExpression ('+', $1, $3, @$); }
| expression '-' expression
  { $$ = new BinaryExpression ('-', $1, $3, @$); }
| expression '*' expression
  { $$ = new BinaryExpression ('*', $1, $3, @$); }
| expression '/' expression
  { $$ = new BinaryExpression ('/', $1, $3, @$); }
| expression '%' expression
  { $$ = new BinaryExpression ('%', $1, $3, @$); }
| expression '^' expression
  { $$ = new BinaryExpression ('^', $1, $3, @$); }
| '-' expression %prec NEG
  { $$ = new UnaryExpression ('-', $2, @$); };

controlexpression:
  NAME
  { $$ = new ArgumentExpression ($1, @$); }
| NAME '[' ranges ']'
  { $$ = new ArgumentExpression ($1, $3, @$); }
| NUM
  { $$ = new NumberExpression ($1, @$); }
| '(' controlexpression ')'
  { $2->SetLocation (@$); $$ = $2; }
| controlexpression '+' controlexpression
  { $$ = new BinaryExpression ('+', $1, $3, @$); }
| controlexpression '-' controlexpression
  { $$ = new BinaryExpression ('-', $1, $3, @$); }
| controlexpression '*' controlexpression
  { $$ = new BinaryExpression ('*', $1, $3, @$); }
| controlexpression '/' controlexpression
  { $$ = new BinaryExpression ('/', $1, $3, @$); }
| controlexpression '%' controlexpression
  { $$ = new BinaryExpression ('%', $1, $3, @$); }
| controlexpression '^' controlexpression
  { $$ = new BinaryExpression ('^', $1, $3, @$); }
| '-' controlexpression %prec NEG
  { $$ = new UnaryExpression ('-', $2, @$); };

expressions:
  expression
  { $$ = $1; }
| vectorexpression
  { $$ = $1;
    $$->SetLocation (@$);
  };

vectorexpression:
  expression ',' expression
  {
    $$ = new VectorExpression ();
    $$->AddExpression ($1);
    $$->AddExpression ($3);
  }
| vectorexpression ',' expression
  {
    $1->AddExpression ($3);
    $$ = $1;
  };

boolean:
  controlexpression '=' controlexpression
  { $$ = new Comparison ('=', $1, $3, @$); }
| controlexpression '>' controlexpression
  { $$ = new Comparison ('>', $1, $3, @$); }
| controlexpression '<' controlexpression
  { $$ = new Comparison ('<', $1, $3, @$); }
| controlexpression GE controlexpression
  { $$ = new Comparison ('g', $1, $3, @$); }
| controlexpression LE controlexpression
  { $$ = new Comparison ('l', $1, $3, @$); }
| boolean '&' boolean
  { $$ = new BooleanBinary ('&', $1, $3, @$); }
| boolean '|' boolean
  { $$ = new BooleanBinary ('|', $1, $3, @$); }
| '!' boolean %prec NOT
  { $$ = new BooleanNot ($2, @$); }
| '(' boolean ')'
  { $$ = $2; };


%%
