#include "argumentexpression.h"
#include "function.h"
#include "rangesexpression.h"
#include "rangeiterator.h"
#include "bivector.h"
#include "instance.h"
#include "yaccmisc.h"
#include <iostream>
#include <stdlib.h>

ArgumentExpression::~ArgumentExpression ()
{ delete Rng; }

AlluvionType ArgumentExpression::Bind (Function &Fun, AlluvionType RequiredType, BindMode BM)
{
  AlluvionType ExistType = Fun.CheckName (Name, &BindVar);
  switch (ExistType)
  {

  case AT_FUNCTION:

    cerr << "ERROR at " << GetLocation ()
         << " in function \"" << Fun.GetName ()
	 << "\": function name \"" << Name
         << "\" used as argument\n";
    exit (1);

  case AT_NO_TYPE:

    if (BM == BM_MUST_EXIST)
    {
      cerr << "ERROR at " << GetLocation ()
           << " in function \"" << Fun.GetName ()
           << "\": argument variable \"" << Name
           << "\" does not exist\n";
      exit (1);
    }

    if (RequiredType == AT_NO_TYPE)
    {
      cerr << "ERROR at " << GetLocation ()
           << " in function \"" << Fun.GetName ()
           << "\": type confusion, this should not have happened!\n";
      exit (1);
    }

    ArgType = RequiredType;
    if (ArgType == AT_VAR && Rng)
       ArgType = AT_VARVECTOR;
    if (ArgType == AT_CONTROLVAR && Rng)
       ArgType = AT_CONTROLVARVECTOR;

    BindVar = Fun.AddNewVariable (Name, ArgType, Rng ? Rng->RangesCount () : 0);

    break;

  default:

    if (BM == BM_MUST_NOT_EXIST)
    {
      cerr << "ERROR at " << GetLocation ()
           << " in function \"" << Fun.GetName ()
           << "\": argument variable \"" << Name
           << "\" already exists\n";
      exit (1);
    }

    if (  RequiredType != AT_NO_TYPE && RequiredType != ExistType
       && ! (RequiredType == AT_VAR && ExistType == AT_VARVECTOR)
       && ! (RequiredType == AT_CONTROLVAR && ExistType == AT_CONTROLVARVECTOR)
       )
    {
      cerr << "ERROR at " << GetLocation ()
           << " in function \"" << Fun.GetName ()
           << "\": argument \"" << Name
           << "\" should be of type " << AlluvionTypeName (RequiredType)
           << ", not " << AlluvionTypeName (ExistType) << endl;
      exit (1);
    }

    if (Rng && ExistType != AT_VARVECTOR && ExistType != AT_CONTROLVARVECTOR)
    {
      cerr << "ERROR at " << GetLocation ()
           << " in function \"" << Fun.GetName ()
	   << "\": " << AlluvionTypeName (ExistType)
           << " \"" << Name
           << "\" may not be indexed\n";
      exit (1);
    }

    if (ExistType == AT_VARVECTOR || ExistType == AT_CONTROLVARVECTOR)
    {
      if (Rng == 0)
      {
        cerr << "ERROR at " << GetLocation ()
             << " in function \"" << Fun.GetName ()
             << "\": " << AlluvionTypeName (ExistType)
             << " \"" << Name
             << "\" must be indexed\n";
        exit (1);
      }
      unsigned IndexDim;
      if   (ExistType == AT_VARVECTOR)
           IndexDim = Fun.GetVarArrayIndexDim (BindVar);
      else IndexDim = Fun.GetControlVarArrayIndexDim (BindVar);
      if (IndexDim != Rng->RangesCount ())
      {
        cerr << "ERROR at " << GetLocation ()
             << " in function \"" << Fun.GetName ()
             << "\": " << AlluvionTypeName (ExistType)
             << " \"" << Name << *Rng
             << "\" should have dimension " << IndexDim << endl;
        exit (1);
      }
    }

    ArgType = ExistType;
  }

  if (Rng) Rng->Bind (Fun);

  switch (ArgType)
  { case AT_VARVECTOR: return AT_VAR;
    case AT_CONTROLVARVECTOR: return AT_CONTROLVAR;
    default: return ArgType;
  }
}

void ArgumentExpression::Evaluate (Instance &In, Instance &Out)
{
  switch (ArgType)
  {
  case AT_VAR:
   {
    Value &Val = In.GetValue (BindVar);
    if (Val.GetState () == ValueNotYet && Out.GetFunction ()->GetSequence () == 0)
    // unset values forbidden for builtin functions
    {
      cerr << "ERROR at " << GetLocation ()
           << " in function \"" << In.GetFunction ()->GetName ()
           << "\": value of variable \""
           << Name << "\" not set\n";
      exit (1);
    }
    Out.WriteValue (Val);
    break;
   }
  case AT_VARVECTOR:
   {
    OffsetArray<Value> &VV = In.GetValueArray (BindVar);
    for (RangesIterator RI (Rng->EvaluateRanges (In)); ! RI.AtEnd (); RI ++)
    {
      Value &Val = VV[*RI];
      if (Val.GetState () == ValueNotYet && Out.GetFunction ()->GetSequence () == 0)
      // unset values forbidden for builtin functions
      {
        cerr << "ERROR at " << GetLocation ()
             << " in function \"" << In.GetFunction ()->GetName ()
	     << "\": value of variable \""
	     << Name << '[' << RI << "]\" not set\n";
        exit (1);
      }
      Out.WriteValue (Val);
    }
    break;
   }
  case AT_CONTROLVAR:
   {
    Out.AccessControlValue () = In.GetControlValue (BindVar);
    break;
   }
  case AT_CONTROLVARVECTOR:
   {
    OffsetArray<int> &CVV = In.GetControlValueArray (BindVar);
    for (RangesIterator RI (Rng->EvaluateRanges (In)); ! RI.AtEnd (); RI ++)
      // TODO: was, wenn nicht vorhanden?
      Out.AccessControlValue () = CVV[*RI];
    break;
   }
  case AT_FUNCTION:
  case AT_NO_TYPE:
    break;
  }
}

int ArgumentExpression::Evaluate (Instance &In)
{
  switch (ArgType)
  {
  case AT_CONTROLVAR:
    return In.GetControlValue (BindVar);
  case AT_CONTROLVARVECTOR:
    {
      vector<Range> &R = *Rng->EvaluateRanges (In);
      vector<int> Indices;
      for (unsigned i = 0; i < R.size (); i ++)
      {
        if (R[i].GetBegin () != R[i].GetEnd ())
        {
          cerr << "ERROR at " << GetLocation ()
               << " in function \"" << In.GetFunction ()->GetName ()
               << "\": indexed control variable should return single control value (i.e. range "
               << R[i] << " too large or empty)\n";
          exit (1);
        }
        Indices.push_back (R[i].GetBegin ());
      }
      delete &R;
      return In.GetControlValueArray (BindVar) [Indices];
    }
  default:
    cerr << "internal ERROR in ArgumentExpression::Evaluate (Instance)\n";
    exit (1);
  }
}

void ArgumentExpression::Print (ostream &s) const
{
  s << Name;
  if (Rng)
    Rng->Print (s);
}

void ArgumentExpression::PrintState (Instance &I)
{
  if (ArgType == AT_CONTROLVAR)
  {
    cout << Name << " = " << I.GetControlValue (BindVar) << endl;
  }
  else if (ArgType == AT_VAR)
  {
    cout << Name << '\t';
    I.GetValue (BindVar).Print ();
  }
  else
  {
    vector<Range> &R = *Rng->EvaluateRanges (I);
    bool single = true;
    for (unsigned i = 0; i < R.size (); i ++)
      if (R[i].GetEnd () != R[i].GetBegin ()) single = false;
    if (!single)
    {
      cout << Name;
      Rng->Print ();
      cout << ":\n";
      for (RangesIterator RI (R); ! RI.AtEnd (); RI ++)
      {
        cout << "  [" << RI << "]\t";
        if   (ArgType == AT_VARVECTOR)
             I.GetValueArray (BindVar) [*RI].Print ();
        else cout << I.GetControlValueArray (BindVar) [*RI] << endl;
      }
    }
    else
    {
      vector<int> Indices;
      cout << Name << '[';
      for (unsigned i = 0; i < R.size (); i ++)
      { if (i > 0) cout << ',';
        cout << R[i].GetBegin ();
        Indices.push_back (R[i].GetBegin ());
      }
      cout << "]\t";
      if   (ArgType == AT_VARVECTOR)
           I.GetValueArray (BindVar) [Indices].Print ();
      else cout << I.GetControlValueArray (BindVar) [Indices] << endl;
    }
    delete &R;
  }
}
